#
# editer.rb
#
#   Copyright (c) 1999 Minero Aoki <aamine@dp.u-netsurf.ne.jp>
#
#   This program is free software.
#   you can distribute/modify this program under the terms of
#   the GNU Lesser General Public License version 2 or later.
#


module EST


  class Editer

    def initialize( port, rc )
      port.must ::TMail::Port

      @port = port
      @port.ref
      @rc = rc
      @shuttingdown = false
      do_initialize
    end

    # for ruby level exit
    def shutdown( force = false )
      return if @shuttingdown
      @shuttingdown = true
      return unless force or shutdown_ok?
      edit_cancel
    end


    private


    def edit_ok
      i = mail2idx( @port )
      CORE.mailbox.draft.select i
      CORE.notice "edit draft\##{i + 1}"
      in_shutdown
    end

    def edit_cancel
      i = mail2idx( @port )
      # CORE.mailbox.draft.remove i
      # CORE.notice "discard draft\##{i + 1}"
      CORE.notice "leave draft\##{i + 1}"
      in_shutdown
    end

    def in_shutdown
      @port.unref
      CORE.edit_done self
    end


    def mail2idx( port )
      CORE.mailbox.draft.index port
    end

  end   # class Editer



  class InternalEditer < Editer

    private

    def do_initialize
      @mail = Mail.new( @port )
      @buf = SizedStack.new

      set_keybind
      construct_gui
      @win.show_all
    end


    def construct_gui
      mrc = @rc['menu']

      @win     = Gtk::Window.new( Gtk::WINDOW_TOPLEVEL )
      @win.set_title( 'mailest editer' )
      @win.border_width(0)
      @win.set_usize( 600, 700 )

      @box     = Gtk::VBox.new( false, 0 )
      @win.add @box

      @butbox  = Gtk::HBox.new( false, 0 )
      @box.pack_start( @butbox, false, false, 0 )

      @finbtn  = Gtk::Button.new( mrc['done'] )
      @finbtn.set_usize( 120, 22 )
      @butbox.pack_start( @finbtn, false, false, 0 )

      @intbtn  = Gtk::Button.new( mrc['cancel'] )
      @intbtn.set_usize( 120, 22 )
      @butbox.pack_start( @intbtn, false, false, 0 )

      @tobox   = Gtk::HBox.new( false, 1 )
      @box.pack_start( @tobox, false, false, 0 )

      @tolab   = Gtk::Label.new( mrc['to'] )
      @tolab.set_usize( 80, 22 )
      @tobox.pack_start( @tolab, false, false, 0 )

      @toent   = Gtk::Entry.new
      @toent.set_usize( 460, 22 )
      @toent.set_editable true
      to = @mail.to(nil)
      to = to ? to.decoded : ''
      @toent.set_text to
      @tobox.pack_start( @toent, true, true, 0 )

      @sbjbox  = Gtk::HBox.new( false, 1 )
      @box.pack_start( @sbjbox, false, false, 0 )

      @sbjlab  = Gtk::Label.new( mrc['subject'] )
      @sbjlab.set_usize( 80, 22 )
      @sbjbox.pack_start( @sbjlab, false, false, 0 )

      @sbjent  = Gtk::Entry.new
      @sbjent.set_usize( 460, 22 )
      @sbjent.set_editable true
      @sbjent.set_text @mail.subject
      @sbjbox.pack_start( @sbjent, true, true, 0 )


      @textbox = Gtk::HBox.new( false, 0 )
      @adj     = Gtk::Adjustment.new(0,0,0,0,0,0)
      @scr     = Gtk::VScrollbar.new( @adj )
      @text    = Gtk::Text.new( nil, @adj )
      # @text    = Gtk::Buffer.new( nil, @adj )
      @text.insert_text( @mail.body, 0 )
      @text.set_editable true

      @textbox.pack_start( @text, true, true, 0 )
      @textbox.pack_start( @scr, false, false, 0 )
      @box.pack_start( @textbox, true, true, 0 )

      set_callback
    end


    def set_callback
      @finbtn.signal_connect( 'clicked' ) do |w|
        protect( 'edit fin' ){
          edit_ok
        }
      end

      @intbtn.signal_connect( 'clicked' ) do |w|
        protect( 'edit cancel' ){
          edit_cancel
        }
      end

      @text.signal_connect( 'key_press_event' ) do |w, eve|
        c = (eve.state & Gdk::CONTROL_MASK != 0)
        a = (eve.state & Gdk::MOD1_MASK != 0)

        bind = if    c and a then @ctrl_alt
               elsif c       then @ctrl
               elsif       a then @alt
               else               @raw
               end
#p eve.keyval
#p bind[ eve.keyval ]
        if blk = bind[ eve.keyval ] then
          emitstop = blk.call( eve )
          @text.signal_emit_stop( 'key_press_event' ) if emitstop
        end
      end
    end


    def set_keybind
      @raw = {}
      @ctrl = {}
      @alt  = {}
      @ctrl_alt = {}

      @ctrl[Gdk::GDK_i] = @ctrl[Gdk::GDK_i] = proc{
        File.open( File.expand_path '~/.signature' ) do |f|
          @text.insert_text f.read, @text.position
        end
        true
      }
      @ctrl[Gdk::GDK_k] = @ctrl[Gdk::GDK_K] = proc{
        pnt = @text.position
        @text.set_point pnt
        if @text.get_chars( pnt, pnt + 1 ) == "\n" then
          @text.forward_delete 1
          true
        else
          false
        end
      }
      @ctrl[Gdk::GDK_s] = @ctrl[Gdk::GDK_S] = proc{
        edit_ok
        false
      }
      @ctrl[Gdk::GDK_q] = @ctrl[Gdk::GDK_Q] = proc{
        edit_cancel
        false
      }
      @ctrl[Gdk::GDK_y] = @ctrl[Gdk::GDK_Y] = proc{
        if cmd = @buf.pop then
          cmd.yank(@text)
        end
        true
      }
    end


    def edit_ok
      @mail.body    = @text.get_chars( 0, @text.get_length )
      @mail.subject = @sbjent.get_text
      @mail.to      = @toent.get_text
      @mail.write_back
      super
    end
  

    def in_shutdown
      @win.hide
      @win.destroy
      super
    end


    def shutdown_ok?
      true
    end

  end


  class SizedStack

    def initialize( up = 10 )
      @upper = up
      @arr = []
    end

    attr :upper, true

    def push( item )
      @arr.push item
      if @arr.size > @upper then
        @arr.shift
      end
    end

    def pop
      @arr.pop
    end

    def each
      @arr.each{|i| yield i }
    end

    def size
      @arr.size
    end

  end


  class Backup
  end

  class DelBackup < Backup
  
    def initialize( from, str )
      @from = from
      @str = str
    end

    def undo( text )
      text.insert_text @str, @from
    end

    def yank( text )
      text.insert_text @str, text.get_point
    end
  
  end

  class MultiBackup < Backup

    def initialize( *args )
      @args = args
    end

    def undo( text )
      @args.each{|cmd| cmd.undo( text ) }
    end

    def yank( text )
      @args.each{|cmd| cmd.yank( text ) }
    end
  
  end


  class ExternalEditer < Editer

    private


    def shutdown_ok?
      true
    end


    def do_initialize
      fname = sprintf( '/tmp/est%x_%x', $$, @port.id )
      File.open( fname, 'w' ) do |f|
        @port.ropen do |rp|
          while tmp = rp.read( 2048 )
            f.write tmp
          end
        end
      end

      Thread.fork do
        do_edit( fname )
      end
    end


    def do_edit( fname )
      stamp = File.mtime( fname )
      system "#{@rc['type']} #{fname}"
      if File.mtime( fname ) == stamp then   # not canceled
        edit_cancel
      else                                   # edited
        @port.wopen do |wp|
          File.open( fname ) do |f|
            while tmp = f.read( 2048 ) do
              wp.write tmp
            end
          end
        end
        edit_ok
      end
    end

  end


end   # module EST
